#####################################################
###   A registry of callable view functions
###   in the DMP system

from django.apps import apps, AppConfig

from .util import get_dmp_instance

import threading


# lock to keep register_app() thread safe
rlock = threading.RLock()

# the DMP-enabled AppConfigs in the system (by name)
DMP_ENABLED_APPS = {}



###############################################################
###   Registration functions for apps and view functions
###

def is_dmp_app(app):
    '''
    Returns True if the given app is a DMP-enabled app.  The app parameter can
    be either the name of the app or an AppConfig object.
    '''
    if app is None:
        return False
    if isinstance(app, AppConfig):
        app = app.name
    return app in DMP_ENABLED_APPS


def register_app(app):
    '''
    Registers an app as a "DMP-enabled" app and creates a cached
    template renderer to make processing faster. The app parameter can
    be either the name of the app or an AppConfig object.

    This is called by MakoTemplates (engine.py) during system startup.
    '''
    if isinstance(app, str):
        app = apps.get_app_config(app)

    # since this only runs at startup, this lock doesn't affect performance
    with rlock:
        # short circuit if already in the DMP_ENABLED_APPS
        if app.name in DMP_ENABLED_APPS:
            return

        # check for the DJANGO_MAKO_PLUS flag in the module
        if not getattr(app.module, 'DJANGO_MAKO_PLUS', False):
            return

        # first time for this app, so add to our dictionary
        DMP_ENABLED_APPS[app.name] = app

        # set up the template, script, and style renderers
        # these create and cache just by accessing them
        get_dmp_instance().get_template_loader(app, 'templates', create=True)
        get_dmp_instance().get_template_loader(app, 'scripts', create=True)
        get_dmp_instance().get_template_loader(app, 'styles', create=True)




from django.utils.encoding import force_text
from django.conf import settings

from ..util import merge_dicts

import os
import os.path



##############################################################
###   Static File Providers

class BaseProvider(object):
    '''
    A list of provider instances is attached to each template in the system.
    During a provider run, all provider lists are run.

        base.htm      [ JsLinkProvider1, CssLinkProvider1, ... ]
           |
        app_base.htm  [ JsLinkProvider2, CssLinkProvider2, ... ]
           |
        index.html    [ JsLinkProvider3, CssLinkProvider3, ... ]

    The data argument sent into provide() spans a run vertically, meaning
    the three JsLinkProviders above share the same data dict.
    '''

    default_options = {
        'group': 'styles',
        'enabled': True,
    }

    def __init__(self, app_config, template_path, options):
        self.app_config = app_config
        self.template_path = template_path
        # this next variable assumes the template is in the "normal" location: app/subdir/
        parts = os.path.splitext(self.template_path)[0].split(os.path.sep)
        self.template_name = os.path.sep.join(parts[1:]) if len(parts) > 1 else self.template_path
        self.options = merge_dicts(self.default_options, options)     # combined options dictionary

    @property
    def group(self):
        return self.options['group']

    def start(self, provider_run, data):
        '''
        Called on the *main* template's provider list as the run starts.
        Initialize values in the data dictionary here.
        '''
        pass

    def provide(self, provider_run, data):
        '''Called on *each* template's provider list in the chain - use provider_run.write() for content'''
        pass

    def finish(self, provider_run, data):
        '''
        Called on the *main* template's provider list as the run finishes
        Finalize values in the data dictionary here.
        '''
        pass



    ##################################
    ###  Helper functions

    def options_format(self, val, **extra):
        '''
        Runs val.format() with some named arguments:

        {appname} - The app name for the template being rendered.
        {template} - The name of the template being rendered, without its extension.
        {appdir} - The app directory for the template being rendered (full path).
        {staticdir} - The static directory as defined in settings.
        '''
        d = {
            'appname': self.app_config.name,
            'appdir': self.app_config.path,
            'template': self.template_name,
            'staticdir': getattr(settings, 'STATIC_ROOT', None),
        }
        d.update(extra)
        return force_text(val).format(**d)






from django.conf import settings
from django.core.management.base import BaseCommand, CommandError
from django_mako_plus.util import DMP_OPTIONS, get_dmp_app_configs

from django_mako_plus.provider import create_mako_context
from django_mako_plus.provider.runner import ProviderRun, create_factories
from django_mako_plus.util import get_dmp_instance, split_app

import glob
import os, os.path, shutil
import json
from collections import OrderedDict



class Command(BaseCommand):
    args = ''
    help = 'Removes compiled template cache folders in your DMP-enabled app directories.'
    can_import_settings = True


    def add_arguments(self, parser):
        parser.add_argument(
            '--verbose',
            action='store_true',
            dest='verbose',
            default=False,
            help='Set verbosity to level 3 (see --verbosity).'
        )
        parser.add_argument(
            '--quiet',
            action='store_true',
            dest='quiet',
            default=False,
            help='Set verbosity to level 0, which silences all messages (see --verbosity).'
        )
        parser.add_argument(
            'appname',
            type=str,
            nargs='*',
            help='The name of an app. If omitted, __entry__.js files are created in all DMP apps.'
        )
        parser.add_argument(
            '--overwrite',
            action='store_true',
            dest='overwrite',
            default=False,
            help='Overwrite existing __entry__.js if necessary.'
        )



    def handle(self, *args, **options):
        # save the options for later
        self.options = options
        if self.options['verbose']:
            self.options['verbosity'] = 3
        if self.options['quiet']:
            self.options['verbosity'] = 0

        # ensure we have a base directory
        try:
            if not os.path.isdir(os.path.abspath(settings.BASE_DIR)):
                raise CommandError('Your settings.py BASE_DIR setting is not a valid directory.  Please check your settings.py file for the BASE_DIR variable.')
        except AttributeError as e:
            print(e)
            raise CommandError('Your settings.py file is missing the BASE_DIR setting.')

        # run for each dmp-enabled app
        self.factories = create_factories('WEBPACK_PROVIDERS')
        for config in get_dmp_app_configs():
            if not options['appname'] or config.name in options['appname']:
                self.create_entry(config)


    def message(self, msg, level=1):
        '''Print a message to the console'''
        # verbosity=1 is the default if not specified in the options
        if self.options['verbosity'] >= level:
            print(msg)


    def create_entry(self, config):
        '''Creates a webpack __entry__.js file in the given app'''
        templates_dir = os.path.join(config.path, 'templates')
        entry_filename = os.path.join(config.path, 'scripts', '__entry__.js')
        # map templates to their scripts
        page_map = OrderedDict()
        for template_name in os.listdir(templates_dir):
            if os.path.isfile(os.path.join(os.path.join(templates_dir, template_name))):
                template_obj = get_dmp_instance().get_template_loader(config, create=True).get_mako_template(template_name)
                pages = self.template_scripts(template_obj)
                page_map.update(pages)
        # write the file
        if not self.options['overwrite'] and os.path.exists(entry_filename):
            raise ValueError('Refusing to destroy existing file: %s.  Use --overwrite option or remove the file.' % (entry_filename,))
        if len(page_map) == 0:
            self.message('Templates in app {} had no matching scripts'.format(config.name))
        else:
            self.message('Templates in app {} required {} script(s); creating {}'.format(config.name, len(page_map), os.path.relpath(entry_filename, settings.BASE_DIR)))
            with open(entry_filename, 'w') as fout:
                fout.write('(context => {\n')
                for page, scripts in page_map.items():
                    fout.write('    DMP_CONTEXT.appBundles["%s"] = () => { %s; };\n' % (
                        page,
                        '; '.join([ 'require("%s")' % (s,) for s in scripts ]),
                    ))
                fout.write('})(DMP_CONTEXT.get());\n')


    def template_scripts(self, template_obj):
        '''Maps the scripts used by the given template and its ancestors'''
        # for this algorithm to work, providers must populate the provider data dictionary like this example.
        # the built-in JsLinkProvider does this already.
        # provider_data = {
        #     'urls': [
        #         '/static/app/scripts/first.js',
        #         '/static/app/scripts/second.js',
        #         ...
        #     ]
        # }
        mako_context = create_mako_context(template_obj)
        inner_run = ProviderRun(mako_context['self'], factories=self.factories)
        inner_run.get_content()
        pages = {}
        for data in inner_run.provider_data:
            # determine the app and template
            app_config, template_path = split_app(template_obj.filename)
            _, filename = template_path.split('/', 1)
            page, _ = os.path.splitext(filename)
            # determine the relative location of the urls
            scripts_dir = os.path.join(app_config.name, 'scripts')
            scripts = []
            for url in data.get('urls', []):
                url = url.split('?')[0]
                scripts.append(os.path.join('.', os.path.relpath(url, scripts_dir)))
            if len(scripts) > 0:
                pages['{}/{}'.format(app_config.name, page)] = scripts
        return pages








from django.conf import settings
from django.utils.module_loading import import_string

from ..util import merge_dicts, getdefaultattr
from ..version import __version__
from .static_links import LinkProvider

import os
import os.path
import json


EMPTY_SET = set()


class AppJsBundleProvider(LinkProvider):
    '''
    Generates a JS <script> for the app-level bundle.

    Special format keywords for use in the options:
        {appname} - The app name for the template being rendered.
        {template} - The name of the template being rendered, without its extension.
        {appdir} - The app directory for the template being rendered (full path).
        {staticdir} - The static directory as defined in settings.
    '''
    default_options = merge_dicts(LinkProvider.default_options, {
        'group': 'scripts',
        'filename': '{appdir}/scripts/__bundle__.js',
        'async': False,
    })

    def start(self, provider_run, data):
        data['scripts'] = []

    def provide(self, provider_run, data):
        # call the function for this view every time
        key = '{}/{}'.format(self.app_config.name, self.template_name)
        data['scripts'].append('if (DMP_CONTEXT.appBundles["%s"]) { DMP_CONTEXT.appBundles["%s"]() };' % (key, key))

    def finish(self, provider_run, data):
        '''Runs only on the main template in the chain.'''
        for fi in self.matches:
            # only print the first instance of a given href (one per request)
            js_done = getdefaultattr(provider_run.request, '__dmp_AppJsBundleProvider__', factory=set) if provider_run.request is not None else EMPTY_SET
            if fi.url not in js_done:
                provider_run.write('<script data-context="{contextid}" src="{static}{href}?{version}"{async}></script>'.format(
                    contextid=provider_run.uid,
                    static=settings.STATIC_URL,
                    href=fi.url,
                    version=provider_run.version_id if provider_run.version_id is not None else fi.mtime,
                    async=' async="async"' if self.options['async'] else '',
                ))
                js_done.add(fi.url)
        # print the script
        provider_run.write('<script data-context="{}">'.format(provider_run.uid))
        for line in data['scripts']:
            provider_run.write(line)
        provider_run.write('</script>')






from django.conf import settings
from django.utils.module_loading import import_string

from ..util import merge_dicts
from ..version import __version__
from .base import BaseProvider

import os
import os.path
import json
import glob
from collections import namedtuple



FileInfo = namedtuple('FileInfo', ( 'url', 'mtime' ))

class LinkProvider(BaseProvider):
    '''
    Renders links like <link> and <script> based on the name of the template
    and supertemplates.

    Special format keywords for use in the options:
        {appname} - The app name for the template being rendered.
        {template} - The name of the template being rendered, without its extension.
        {appdir} - The app directory for the template being rendered (full path).
        {staticdir} - The static directory as defined in settings.
    '''
    # the default options are for CSS files
    default_options = merge_dicts(BaseProvider.default_options, {
        # subclasses override both of these
        'group': 'static.file',
        'filename': '{appdir}/somedir/{template}.static.file',
    })

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.matches = []
        for fullpath in glob.iglob(self.options_format(self.options['filename'])):
            self.matches.append(FileInfo(
                os.path.relpath(fullpath, settings.BASE_DIR).replace('\\', '/'),
                int(os.stat(fullpath).st_mtime),
            ))


class CssLinkProvider(LinkProvider):
    '''Generates a CSS <link>'''
    default_options = merge_dicts(LinkProvider.default_options, {
        'group': 'styles',
        'filename': '{appdir}/styles/{template}.css',
        'skip_duplicates': True,
    })

    def provide(self, provider_run, data):
        for fi in self.matches:
            provider_run.write('<link data-context="{contextid}" rel="stylesheet" type="text/css" href="{static}{url}?{version}" />'.format(
                contextid=provider_run.uid,
                static=settings.STATIC_URL,
                url=fi.url,
                version=provider_run.version_id if provider_run.version_id is not None else fi.mtime,
                skip_duplicates='true' if self.options['skip_duplicates'] else 'false',
            ))


class JsLinkProvider(LinkProvider):
    '''Generates a JS <script>.'''
    default_options = merge_dicts(LinkProvider.default_options, {
        'group': 'scripts',
        'filename': '{appdir}/scripts/{template}.js',
        'async': False,
    })

    def start(self, provider_run, data):
        data['urls'] = []

    def provide(self, provider_run, data):
        # delaying printing of tag because the JsContextProvider delays and this must go after it
        for fi in self.matches:
            data['urls'].append('{}?{}'.format(
                fi.url.replace('\\', '/'),
                provider_run.version_id if provider_run.version_id is not None else fi.mtime,
            ))

    def finish(self, provider_run, data):
        for url in data['urls']:
            provider_run.write('<script data-context="{contextid}" src="{static}{url}"{async}></script>'.format(
                contextid=provider_run.uid,
                static=settings.STATIC_URL,
                url=url,
                async=' async="async"' if self.options['async'] else '',
            ))


class JsContextProvider(BaseProvider):
    '''
    Adds all js_context() variables to DMP_CONTEXT.
    This should be listed before JsLinkProvider so the
    context variables are available during <script> runs.
    '''
    default_options = merge_dicts(LinkProvider.default_options, {
        'group': 'scripts',
        'encoder': 'django.core.serializers.json.DjangoJSONEncoder',
    })

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.encoder = import_string(self.options['encoder'])
        self.template = "{}/{}".format(self.app_config.name, self.template_name)

    def start(self, provider_run, data):
        data['templates'] = []

    def provide(self, provider_run, data):
        data['templates'].append(self.template)

    def finish(self, provider_run, data):
        if len(data['templates']) == 0:
            return
        context_data = {
            jscontext('__router__'): {
                'template': self.template,
                'app': provider_run.request.dmp.app if provider_run.request is not None else None,
                'page': provider_run.request.dmp.page if provider_run.request is not None else None,
            },
        }
        context_data.update({ k: provider_run.context[k] for k in provider_run.context.kwargs if isinstance(k, jscontext) })
        provider_run.write('<script>')
        provider_run.write('DMP_CONTEXT.set("{version}", "{contextid}", {data}, {templates});'.format(
            version=__version__,
            contextid=provider_run.uid,
            data=json.dumps(context_data, cls=self.encoder, separators=(',', ':')) if context_data else '{}',
            templates=json.dumps(data['templates']),
        ))
        provider_run.write('</script>')


class jscontext(str):
    '''
    Marks a key in the context dictionary as a JS context item.
    JS context items are sent to the template like normal,
    but they are also added to the runtime JS namespace.

    See the tutorial for more information on this function.
    '''
    # no code needed, just using the class for identity

